// ////////////////////////////////////////////////////
//
// StdView.cpp : implementation of the CStdView class
// Based on class COGLView : public CView
// Inspect COGLView for standard OpenGL initialization
//
////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Std.h"

#include "StdDoc.h"
#include "StdView.h"

#include <math.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////
// You can put definitions, variables and helperfunctions
// here. Functions that does not rely on CStdView member
// variables or functions
#define NO_OBJECT -99	// signal no object hilited
#define BUFSIZE 1024	// store hits


#define PI 3.1415f

#define TRAIL_R 30.0f	// trails semicircle radius
#define TRAIL_L 50.0f	// trails length of sides
#define TRAIL_B 1.0f	// trailwidth

#define REF_X -100.0f	// reference point for trail
#define REF_Y 100.0f	// reference point
#define REF_Z 0.0f		// reference point


void DrawCart(float s)
{
	// s is length from startpos
	// startpnt is at start of north length
	
	glPushMatrix();

	// get in reference position
	glTranslatef(REF_X,REF_Y,REF_Z);

	// will not deal with neg distance
	if(s<0.0f)
		s=-s;
	
	// reduce length to part of one round
	while(s > 2.0f*(PI*TRAIL_R+TRAIL_L))
		s-=2.0f*(PI*TRAIL_R+TRAIL_L);

	// clockwise
	s=2.0f*(PI*TRAIL_R+TRAIL_L)-s;

	// which part of trail?
	if(s<TRAIL_L)
	{
		// north length
		glTranslatef(TRAIL_L-s,TRAIL_R,0.0f);
	}
	else if(s < TRAIL_L+PI*TRAIL_R)
	{
		// left curve
		s-=TRAIL_L; // length in curve
		float v=PI*(0.5f+s/(PI*TRAIL_R)); // as angle
		glTranslatef(TRAIL_R*(float)cos(v),TRAIL_R*(float)sin(v),0.0f);
	}
	else if(s < 2*TRAIL_L+PI*TRAIL_R)
	{
		// south length
		s-=TRAIL_L+PI*TRAIL_R; // along length
		glTranslatef(s,-TRAIL_R,0.0f);
	}
	else
	{
		// right curve
		s-=2*TRAIL_L+PI*TRAIL_R; // length in curve
		float v=PI*(1.5f+s/(PI*TRAIL_R)); // as angle
		glTranslatef(TRAIL_L+TRAIL_R*(float)cos(v),TRAIL_R*(float)sin(v),0.0f);
	}

	// draw the ball
	glTranslatef(0.0f,0.0f,TRAIL_B);
	CMaterial::SetMaterial(RED_PLASTIC);
	GLUquadricObj *glpQ;
	glpQ=gluNewQuadric();
	gluSphere(glpQ,2*TRAIL_B,20,20);
	gluDeleteQuadric(glpQ);

	glPopMatrix();
}

void DrawTunnel()
{
	// draw a tunnel covering half of north length
	glPushMatrix();
	glTranslatef(REF_X,REF_Y,REF_Z);
	
	// set up clipping plane
	GLdouble cutplane[4]={0.0f,0.0f,1.0f,0.0f};
	glClipPlane(GL_CLIP_PLANE2,cutplane);
	glEnable(GL_CLIP_PLANE2);
	
	glTranslatef(0.0f,TRAIL_R,0.0f);
	glRotatef(90.0f,0.0f,1.0f,0.0f);
	glTranslatef(0.0f,0.0f,TRAIL_L/4.0f);
	
	
	CMaterial::SetMaterial(EMERALD);
	GLUquadricObj *glpQ;
	glpQ=gluNewQuadric();
	gluCylinder(glpQ,5*TRAIL_B,5*TRAIL_B,TRAIL_L/2.0f,20,20);
	gluDeleteQuadric(glpQ);
	
	glDisable(GL_CLIP_PLANE2);
	glPopMatrix();
}

void DrawTrail()
{
	// get into position here
	// since this function should be implanted
	// in a bigger and unknown environment
	glPushMatrix();
	glTranslatef(REF_X,REF_Y,REF_Z);
	// we are at the trails "origo"
	glLineWidth(2.0f); // fat track
	
	// set up some circle increments
	float cv=(float)cos(2*PI/100.0);
	float sv=(float)sin(2*PI/100.0);

	// do inner track
	float x=0.0f;
	float y=TRAIL_R;
	glBegin(GL_LINE_LOOP);
		glVertex3f(x,y,0.0f);
		for(int ix=0;ix<100;ix++)
		{
			float xt=x*cv-y*sv;
			float yt=x*sv+y*cv;
			if(ix>50)
				glVertex3f(xt+TRAIL_L,yt,0.0f);// right semicircle
			else
				glVertex3f(xt,yt,0.0f); // left semicircle
			x=xt; y=yt;
		}
	glEnd();

	// do outer track, same strategy
	x=0.0f;
	y=TRAIL_R+TRAIL_B;
	glBegin(GL_LINE_LOOP);
		glVertex3f(x,y,0.0f);
		for(int ix=0;ix<100;ix++)
		{
			float xt=x*cv-y*sv;
			float yt=x*sv+y*cv;
			if(ix>50)
				glVertex3f(xt+TRAIL_L,yt,0.0f);// right semicircle
			else
				glVertex3f(xt,yt,0.0f); // left (low x) semicircle
			x=xt; y=yt;
		}
	glEnd();

	glPopMatrix();
}

void DrawHelpLines()
{
	
	glPushMatrix();

	glTranslatef(REF_X,REF_Y,0.0f);
	glLineWidth(1.0f);
	
	glBegin(GL_LINES);
	glVertex3f(0.0f,-1.25f*TRAIL_R,0.0f);
	glVertex3f(0.0f,1.25f*TRAIL_R,0.0f);

	glVertex3f(TRAIL_L,-1.25f*TRAIL_R,0.0f);
	glVertex3f(TRAIL_L,+1.25f*TRAIL_R,0.0f);

	glVertex3f(0.0f,0.0f,0.0f);
	glVertex3f(TRAIL_L,0.0f,0.0f);

	glEnd();

	glPopMatrix();
}



//////////////////////////////////////////////////////////
// CStdView

IMPLEMENT_DYNCREATE(CStdView, COGLView)

BEGIN_MESSAGE_MAP(CStdView, COGLView)
	//{{AFX_MSG_MAP(CStdView)
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_RBUTTONDOWN()
	ON_COMMAND(ID_DRAW_LINES, OnDrawLines)
	ON_UPDATE_COMMAND_UI(ID_DRAW_LINES, OnUpdateDrawLines)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

///////////////////////////////////////////////////////////
// CStdView construction/destruction

CStdView::CStdView()
{
	m_xv=0.0f;
	m_zv=0.0f;
	m_s=0.0f;
	m_DrawLines=FALSE;
	m_hRC=NULL;
	m_hDC=NULL;


};

CStdView::~CStdView()
{}

////////////////////////////////////////////////////////////
// CStdView diagnostics
// You should usually not touch these:
#ifdef _DEBUG
void CStdView::AssertValid() const
{COGLView::AssertValid();}
void CStdView::Dump(CDumpContext& dc) const
{COGLView::Dump(dc);}

CStdDoc* CStdView::GetDocument() // non-debug version is inline
{	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CStdDoc)));
	return (CStdDoc*)m_pDocument;
}
#endif //_DEBUG

//////////////////////////////////////////////////////////////
// CStdView operations

void CStdView::DrawScene()
{
	// Clear the color and depth buffers.
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		
	// Set up for scene
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// position the eye relative to scene
	gluLookAt(	0.0f,0.0f,50.0f,	// eye
				REF_X+TRAIL_L/2.0f,REF_Y,REF_Z,	// looking at
				0.0f,0.0f,1.0f);	// z is up
	
	// rotate around z-axis 
	glRotatef(m_zv,0.0f,0.0f,1.0f);
	// rotate around x-axis 
	glRotatef(m_xv,1.0f,0.0f,0.0f);
	
	glLoadName(NO_OBJECT);
	// draw trail
	DrawTrail();
	// draw Tunnel
	DrawTunnel();
	// draw ball
	glLoadName(1);
	DrawCart(m_s);
	glLoadName(NO_OBJECT);

	// draw help lines
	if(m_DrawLines)
		DrawHelpLines();

	// finnish drawing
	glFlush();
}


void CStdView::InitializeRContext()
{
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    
	// prepare ligthsource
	GLfloat ambient[] = {0.2f,0.2f,0.2f,1.0f };
    GLfloat diffuse[] = {1.0f,1.0f,1.0f,1.0f };
    GLfloat position[] = {20.0f,30.0f,10.0f,0.0f };
    GLfloat lmodel_ambient[] = {0.4f,0.4f,0.4f,1.0f };


    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, position);
	
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);
	
	// smooth the drawing
	glShadeModel(GL_SMOOTH);

    // set background to white
	glClearColor(1.0f, 1.0f, 1.0f, 0.0f);

	glEnable(GL_LINE_SMOOTH);
}

void CStdView::SetPerspective(int cx, int cy)
{
    // We will work on the projection matrix,
	// so set matrix mode
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	// set up your viewing volume
	gluPerspective(40.0f, (GLdouble)cx/cy, 0.1f, 1000.0f);
	// reset matrix mode 
	glMatrixMode(GL_MODELVIEW);
}


void CStdView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// User has pressed left button.Make a note
	// of this and prepare for intuitive rotation of scene
	m_PLast=point;
	SetCapture();
	CView::OnLButtonDown(nFlags, point);
}

void CStdView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	// free capture and terminate rotation
	ReleaseCapture();
	CView::OnLButtonUp(nFlags, point);
}

void CStdView::OnMouseMove(UINT nFlags, CPoint point) 
{
	// if left button is down 
	if(nFlags & MK_LBUTTON)
	{	
		// and has moved a significant dist
		if((abs(m_PLast.x-point.x)>2)||(abs(m_PLast.y-point.y)>2))
		{
			// then we find an intuitive rot angle
			CRect R;
			GetClientRect(&R);
			
			m_zv+=90.0f*((GLfloat (point.x-m_PLast.x))/GLfloat (R.Width()));
			m_xv+=90.0f*((GLfloat (point.y-m_PLast.y))/GLfloat (R.Height()));
			
			// rememeber point and force immediate redraw 
			m_PLast=point;
			RedrawWindow();
		}
	}
}


void CStdView::OnRButtonDown(UINT nFlags, CPoint point) 
{
	// User has pressed right button
	// either start ball or stop ball
	// start it if we have hit it
	int hit=PickObject(point.x,point.y);
	if(hit==1)
	{
		MSG msg;
		do
		{
			m_s+=TRAIL_L/30.0f;
			if(m_s > 2.0f*(PI*TRAIL_R+TRAIL_L))
				m_s-=2.0f*(PI*TRAIL_R+TRAIL_L);
			RedrawWindow();
		}while (!PeekMessage(&msg,GetSafeHwnd(),
					WM_RBUTTONDOWN,WM_RBUTTONDOWN,
					PM_REMOVE));

	}
}


int CStdView::PickObject(int x, int y)
{
	// called when mousedown to identify an object, if any
	
	// must set proper rendering context
	wglMakeCurrent(m_hDC,m_hRC);
	GLuint selectBuf[BUFSIZE]; // to store OGLs reports
	GLint hits;
	GLint viewport[4];// to store current viewport

	glGetIntegerv(GL_VIEWPORT,viewport);// get viewport
	glSelectBuffer(BUFSIZE,selectBuf);// tell OGL where to report
	glRenderMode(GL_SELECT);// as opposed to GL_RENDER

	glInitNames();
	glPushName(99); // since -1 gives a warning

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	// set up a small clipping pyramid
	gluPickMatrix(	(GLdouble)x,(GLdouble)(viewport[3]-y),//center
					5.0f,5.0f, // extension
					viewport);

	// same view as we see it,
	// same as in OnSize
	CRect R;
	GetClientRect(&R);
	gluPerspective(	40.0f,(GLdouble)R.Width()/(GLdouble)R.Height(),
					0.1f,1000.0f);
	glMatrixMode(GL_MODELVIEW);

	m_Drawing_Mode=GL_SELECT;
	DrawScene();// draw without rendering
	m_Drawing_Mode=GL_RENDER;

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	
	hits=glRenderMode(GL_RENDER);// gives me a hitcount
	
	// process the ix+1 hits
	// selectBuf contents:
	//	ix  : number of names on the hitstack when hit occurred
	//	ix+1: min z-depth of "object" associated with name, relative
	//	ix+2: max z-depth of "object" associated with name, relative
	//	ix+3: the name, ie index, assosiated with the hit
	//	etc
	if(hits ==0)
		return NO_OBJECT;
	else
	{
		// we have hit one or more "objects"
		// if only one we simply select it
		if(hits ==1)
			return selectBuf[3];//the only one in store

		// more than one is hit.  Which should we select ?
		// The closest one, 
		// go for the one which closest part is closest
		GLint closestix=0;
		for(GLint ix=1;ix<hits;ix++)
			if(selectBuf[4*ix+1]<selectBuf[4*closestix+1])
				closestix=ix;
		return selectBuf[4*closestix+3];
	}
	wglMakeCurrent(0,0);

}


void CStdView::OnDrawLines() 
{
	// drawing help lines
	// for illustration purpose
	m_DrawLines=!m_DrawLines;
	RedrawWindow();
	
}

void CStdView::OnUpdateDrawLines(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_DrawLines);
	
}
